import jwt
from jwt import PyJWTError
from rest_framework.authentication import BaseAuthentication
from rest_framework.exceptions import AuthenticationFailed

from drfjwtauth.settings import ALGORITHM, SIGNING_KEY, AUTH_HEADER, EXPIRATION_DATE_FIELD, QUERY_PARAM_NAME
from drfjwtauth.user import JWTUser


class JWTAuth(BaseAuthentication):

    @classmethod
    def encode_token(cls, payload):
        return jwt.encode(payload, SIGNING_KEY, algorithm=ALGORITHM)

    @classmethod
    def decode_token(cls, encoded_token):
        return jwt.decode(encoded_token, SIGNING_KEY, algorithms=ALGORITHM)

    @classmethod
    def get_token_from_query_params(cls, request):
        return request.query_params.get(QUERY_PARAM_NAME)

    @classmethod
    def get_token_from_auth_header(cls, request):
        token = None
        header_value = request.META.get('HTTP_AUTHORIZATION')
        if header_value:
            if AUTH_HEADER not in header_value:
                pass  # TODO: launch error
            parts = header_value.split(AUTH_HEADER)
            token = parts[1].strip()
        return token

    @classmethod
    def get_token_from_request(cls, request):
        encoded_token = cls.get_token_from_query_params(request)
        if not encoded_token:
            encoded_token = cls.get_token_from_auth_header(request)
    
        return encoded_token

    def authenticate(self, request):
        user = None
        encoded_token = self.get_token_from_request(request)

        if encoded_token:
            try:
                data = self.decode_token(encoded_token)
            except PyJWTError as e:
                error = str(e)
                raise AuthenticationFailed(f'Invalid token: {error}')
            user = JWTUser(data)
        return user, None



from jwt import PyJWTError
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from drfjwtauth.auth import JWTAuth


class VerifyJWTView(APIView):

    def post(self, request):
        token = request.data.get('token') or JWTAuth.get_token_from_request(request)

        # Accept JWT as token parameter in the body and in the Authorization header as 
        # authentication class does.
        if not token:
            return Response({
                'error': 'Missing JWT in the received request. Set it in the body or in the Authentication header'
                },
                status=status.HTTP_400_BAD_REQUEST
            )

        try:
            JWTAuth.decode_token(token)
        except PyJWTError as e:
            return Response({'error': str(e)}, status=status.HTTP_400_BAD_REQUEST)
        return Response({'message': 'ok'})




class JWTUser:
    pk = None
    first_name = ''
    last_name = ''
    username = ''
    email = ''
    is_staff = False
    is_active = False
    is_superuser = False
    channels = []
    groups = []
    permissions = []

    def __str__(self):
        return self.username

    def __int__(self):
        raise self.pk

    def __init__(self, payload):
        self.pk = payload.get('pk', payload.get('id'))
        self.first_name = payload.get('first_name')
        self.last_name = payload.get('last_name')
        self.username = payload.get('username')
        self.email = payload.get('email')
        self.is_staff = payload.get('is_staff')
        self.is_active = payload.get('is_active')
        self.is_superuser = payload.get('is_superuser')
        self.channels = payload.get('channels', [])
        self.groups = payload.get('groups', [])
        self.permissions = payload.get('permissions', [])

    def save(self):
        raise NotImplementedError("Django doesn't provide a DB representation for AnonymousUser.")

    def delete(self):
        raise NotImplementedError("Django doesn't provide a DB representation for AnonymousUser.")

    def set_password(self, raw_password):
        raise NotImplementedError("Django doesn't provide a DB representation for AnonymousUser.")

    def check_password(self, raw_password):
        raise NotImplementedError("Django doesn't provide a DB representation for AnonymousUser.")

    @property
    def id(self):
        return self.pk

    @property
    def user_permissions(self):
        return self.permissions

    def get_user_permissions(self, obj=None):
        return self.permissions

    def get_group_permissions(self, obj=None):
        raise NotImplemented()

    def get_all_permissions(self, obj=None):
        return self.permissions

    def has_perm(self, perm, obj=None):
        return perm in self.permissions

    def has_perms(self, perm_list, obj=None):
        return all(self.has_perm(perm, obj) for perm in perm_list)

    @property
    def is_anonymous(self):
        return False

    @property
    def is_authenticated(self):
        return True

    def get_username(self):
        return self.username



from django.conf import settings

DRF_JWT_AUTH = getattr(settings, 'DRF_JWT_AUTH', {})

ALGORITHM = DRF_JWT_AUTH.get('ALGORITHM', 'HS256')
SIGNING_KEY = DRF_JWT_AUTH.get('SIGNING_KEY', settings.SECRET_KEY)
AUTH_HEADER = DRF_JWT_AUTH.get('AUTH_HEADER', 'Bearer')
QUERY_PARAM_NAME = DRF_JWT_AUTH.get('QUERY_PARAM_NAME', 'jwt')
EXPIRATION_DATE_FIELD = DRF_JWT_AUTH.get('EXPIRATION_DATE_FIELD', 'exp')



from datetime import datetime, timedelta
from unittest import TestCase

from rest_framework import status
from rest_framework.exceptions import AuthenticationFailed

from drfjwtauth import settings
from drfjwtauth.auth import JWTAuth
from drfjwtauth.views import VerifyJWTView


class Request:

    def __init__(self):
        self.META = dict()
        self.data = dict()
        self.query_params = dict()


class JWTTestsUtils(object):

    def setUp(self):
        self.payload = {
            'id': 1,
            'fist_name': 'Name',
            'last_name': 'Company',
            'username': 'ncompany',
            'email': 'name@company.com',
            'is_active': True,
            'is_staff': True,
            'is_superuser': False,
            'channels': ['c1', 'c2', 'c3', 'c4'],
            'groups': ['g1', 'g2'],
            'permissions': ['p1', 'p2', 'p3'],
            settings.EXPIRATION_DATE_FIELD: int((datetime.utcnow() + timedelta(days=1)).timestamp())
        }
        settings.SIGNING_KEY = 'qwertyuiop1234567890'

    def _generate_jwt(self, expired=False):
        if expired:
            self.payload[settings.EXPIRATION_DATE_FIELD] = int((datetime.utcnow() - timedelta(days=1)).timestamp())
        return JWTAuth.encode_token(self.payload).decode('utf-8')

    def _get_request_with_token_in_headers(self, token):
        request = Request()
        request.META['HTTP_AUTHORIZATION'] = f'{settings.AUTH_HEADER} {token}'
        return request


class TestJWTAuth(JWTTestsUtils, TestCase):

    def _get_request_with_token_as_query_param(self, token):
        request = Request()
        request.query_params[settings.QUERY_PARAM_NAME] = token
        return request

    def _authenticate_request(self, request):
        backend = JWTAuth()
        return backend.authenticate(request)

    def _check_user_data_is_properly_loaded(self, user):
        self.assertEqual(user.id, self.payload.get('id'))
        self.assertEqual(user.pk, self.payload.get('id'))
        self.assertEqual(user.first_name, self.payload.get('first_name'))
        self.assertEqual(user.last_name, self.payload.get('last_name'))
        self.assertEqual(user.username, self.payload.get('username'))
        self.assertEqual(user.email, self.payload.get('email'))
        self.assertEqual(user.is_active, self.payload.get('is_active'))
        self.assertEqual(user.is_staff, self.payload.get('is_staff'))
        self.assertEqual(user.is_superuser, self.payload.get('is_superuser'))
        self.assertEqual(user.channels, self.payload.get('channels'))
        self.assertEqual(user.groups, self.payload.get('groups'))
        self.assertEqual(user.permissions, self.payload.get('permissions'))

    def test_invalid_token_in_headers_raises_exception(self):
        request = self._get_request_with_token_in_headers('abc')
        with self.assertRaises(AuthenticationFailed):
            self._authenticate_request(request)

    def test_modified_token_in_headers_raises_exception(self):
        encoded_jwt = self._generate_jwt().replace('e', 'i')
        request = self._get_request_with_token_in_headers(encoded_jwt)
        with self.assertRaises(AuthenticationFailed):
            self._authenticate_request(request)

    def test_expired_token_in_headers_raises_exception(self):
        encoded_jwt = self._generate_jwt(expired=True)
        request = self._get_request_with_token_in_headers(encoded_jwt)
        with self.assertRaises(AuthenticationFailed):
            self._authenticate_request(request)

    def test_user_authenticates_in_headers_and_gets_data_loaded_properly(self):
        encoded_jwt = self._generate_jwt()
        request = self._get_request_with_token_in_headers(encoded_jwt)
        user, unused = self._authenticate_request(request)
        self._check_user_data_is_properly_loaded(user)

    def test_invalid_token_as_query_param_raises_exception(self):
        request = self._get_request_with_token_as_query_param('abc')
        with self.assertRaises(AuthenticationFailed):
            self._authenticate_request(request)

    def test_modified_token_as_query_param_raises_exception(self):
        encoded_jwt = self._generate_jwt().replace('e', 'i')
        request = self._get_request_with_token_as_query_param(encoded_jwt)
        with self.assertRaises(AuthenticationFailed):
            self._authenticate_request(request)

    def test_expired_token_as_query_param_raises_exception(self):
        encoded_jwt = self._generate_jwt(expired=True)
        request = self._get_request_with_token_as_query_param(encoded_jwt)
        with self.assertRaises(AuthenticationFailed):
            self._authenticate_request(request)

    def test_user_authenticates_as_query_param_and_gets_data_loaded_properly(self):
        encoded_jwt = self._generate_jwt()
        request = self._get_request_with_token_as_query_param(encoded_jwt)
        user, unused = self._authenticate_request(request)
        self._check_user_data_is_properly_loaded(user)


class TestVerifyJWTView(JWTTestsUtils, TestCase):

    def _get_request_with_token_data(self, token):
        request = Request()
        request.data['token'] = token
        return request

    def _make_request(self, request):
        view = VerifyJWTView()
        return view.post(request)

    def test_invalid_token_returns_400_bad_request(self):
        request = self._get_request_with_token_data('abc')
        response = self._make_request(request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_modified_token_returns_400_bad_request(self):
        encoded_jwt = self._generate_jwt().replace('e', 'i')
        request = self._get_request_with_token_data(encoded_jwt)
        response = self._make_request(request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_expired_token_returns_400_bad_request(self):
        encoded_jwt = self._generate_jwt(expired=True)
        request = self._get_request_with_token_data(encoded_jwt)
        response = self._make_request(request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_valid_token_returns_200_ok(self):
        encoded_jwt = self._generate_jwt()
        request = self._get_request_with_token_data(encoded_jwt)
        response = self._make_request(request)
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_request_without_token_returns_400_bad_request(self):
        request = Request()
        response = self._make_request(request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_valid_token_in_headers_returns_200_ok(self):
        encoded_jwt = self._generate_jwt()
        request = self._get_request_with_token_in_headers(encoded_jwt)
        response = self._make_request(request)
        self.assertEqual(response.status_code, status.HTTP_200_OK)
